'use strict';

var _slicedToArray = (function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; })();

/**
 * # Typecheck Transformer
 */

Object.defineProperty(exports, "__esModule", {
  value: true
});

exports.default = function (_ref) {
  var t = _ref.types;
  var template = _ref.template;

  // constants used when statically verifying types
  var TYPE_INVALID = 0;
  var TYPE_VALID = 1;
  var TYPE_UNKNOWN = 2;

  /**
   * Binary Operators that can only produce boolean results.
   */
  var BOOLEAN_BINARY_OPERATORS = ['==', '===', '>=', '<=', '>', '<', 'instanceof'];

  var checks = createChecks();
  var staticChecks = createStaticChecks();

  var checkIsArray = template('Array.isArray(input)');
  var checkIsMap = template('input instanceof Map');
  var checkIsSet = template('input instanceof Set');
  var checkIsObject = template('input != null && typeof input === \'object\'');
  var checkNotNull = template('input != null');

  var declareTypeChecker = template('\n    const id = function id (input) {\n      return check;\n    };\n  ');

  var guard = template('\n    if (!check) {\n      throw new TypeError(message);\n    }\n  ');

  var thrower = template('\n    if (check) {\n      ret;\n    }\n    else {\n      throw new TypeError(message);\n    }\n  ');

  var readableName = template('\n    input === null ? \'null\' : typeof input === \'object\' && input.constructor ? input.constructor.name || \'[Unknown Object]\' : typeof input\n  ');

  var checkMapKeys = template('\n    input instanceof Map && Array.from(input.keys()).every(key => keyCheck)\n  ');

  var checkMapValues = template('\n    input instanceof Map && Array.from(input.values()).every(value => valueCheck)\n  ');

  var checkMapEntries = template('\n    input instanceof Map && Array.from(input).every(([key, value]) => keyCheck && valueCheck)\n  ');

  var checkSetEntries = template('\n    input instanceof Set && Array.from(input).every(value => valueCheck)\n  ');

  var stack = [];

  return {
    inherits: require("babel-plugin-syntax-flow"),
    visitor: {
      TypeAlias: function TypeAlias(path) {
        path.replaceWith(createTypeAliasChecks(path));
      },
      InterfaceDeclaration: function InterfaceDeclaration(path) {
        path.replaceWith(createInterfaceChecks(path));
      },
      ExportNamedDeclaration: function ExportNamedDeclaration(path) {
        var node = path.node;
        var scope = path.scope;

        if (node.declaration.type === 'TypeAlias') {
          path.replaceWith(t.exportNamedDeclaration(createTypeAliasChecks(path.get('declaration')), [], null));
        }
      },
      ImportDeclaration: function ImportDeclaration(path) {
        if (path.node.importKind !== 'type') {
          return;
        }

        var _path$get$map$reduce = path.get('specifiers').map(function (specifier) {
          var local = specifier.get('local');
          var tmpId = path.scope.generateUidIdentifierBasedOnNode(local.node);
          var replacement = t.importSpecifier(tmpId, specifier.node.imported);
          var id = t.identifier(local.node.name);

          id.isTypeChecker = true;
          var declarator = t.variableDeclarator(id, tmpId);
          declarator.isTypeChecker = true;
          return [declarator, replacement];
        }).reduce(function (_ref2, _ref3) {
          var _ref5 = _slicedToArray(_ref2, 2);

          var declarators = _ref5[0];
          var specifiers = _ref5[1];

          var _ref4 = _slicedToArray(_ref3, 2);

          var declarator = _ref4[0];
          var specifier = _ref4[1];

          declarators.push(declarator);
          specifiers.push(specifier);
          return [declarators, specifiers];
        }, [[], []]);

        var _path$get$map$reduce2 = _slicedToArray(_path$get$map$reduce, 2);

        var declarators = _path$get$map$reduce2[0];
        var specifiers = _path$get$map$reduce2[1];

        var declaration = t.variableDeclaration('var', declarators);
        declaration.isTypeChecker = true;

        path.replaceWithMultiple([t.importDeclaration(specifiers, path.node.source), declaration]);
      },

      Function: {
        enter: function enter(path) {
          var _node$body$body;

          var node = path.node;
          var scope = path.scope;

          var paramChecks = collectParamChecks(path);
          if (node.type === "ArrowFunctionExpression" && node.expression) {
            node.expression = false;
            node.body = t.blockStatement([t.returnStatement(node.body)]);
          }
          (_node$body$body = node.body.body).unshift.apply(_node$body$body, paramChecks);
          var isVoid = node.returnType ? maybeNullableAnnotation(node.returnType) : null;
          node.savedTypeAnnotation = node.returnType;
          stack.push({ node: node, returns: 0, isVoid: isVoid, type: node.returnType });
        },
        exit: function exit(path) {
          var _stack$pop = stack.pop();

          var node = _stack$pop.node;
          var returns = _stack$pop.returns;
          var isVoid = _stack$pop.isVoid;
          var type = _stack$pop.type;

          if (isVoid === false && returns === 0) {
            throw path.buildCodeFrameError('Function ' + (node.id ? '"' + node.id.name + '" ' : '') + 'did not return a value, expected ' + humanReadableType(type, path.scope));
          }
        }
      },

      ReturnStatement: function ReturnStatement(path) {
        var node = path.node;
        var parent = path.parent;
        var scope = path.scope;

        if (stack.length === 0 || node.isTypeChecked) {
          return;
        }
        stack[stack.length - 1].returns++;
        var fn = stack[stack.length - 1].node;
        var returnType = fn.returnType;

        if (!returnType) {
          return;
        }
        if (!node.argument) {
          if (maybeNullableAnnotation(returnType) === false) {
            throw path.buildCodeFrameError('Function ' + (fn.id ? '"' + fn.id.name + '" ' : '') + 'did not return a value, expected ' + humanReadableType(returnType, path.scope));
          }
          return;
        }
        var id = undefined;
        if (node.argument.type === 'Identifier') {
          id = node.argument;
        } else {
          id = scope.generateUidIdentifierBasedOnNode(node.argument);
        }
        var ok = staticCheckAnnotation(path.get("argument"), returnType);
        if (ok === true) {
          return;
        } else if (ok === false) {
          throw path.buildCodeFrameError('Invalid return type, expected ' + humanReadableType(returnType, scope));
        }
        var check = checkAnnotation(id, returnType, scope);
        if (!check) {
          return;
        }
        if (parent.type !== 'BlockStatement' && parent.type !== 'Program') {
          var block = [];
          if (node.argument.type !== 'Identifier') {
            scope.push({ id: id });
            block.push(t.expressionStatement(t.assignmentExpression('=', id, node.argument)));
          }
          var ret = t.returnStatement(id);
          ret.isTypeChecked = true;
          block.push(thrower({
            check: check,
            ret: ret,
            message: returnTypeErrorMessage(path, fn, id)
          }));
          path.replaceWith(t.blockStatement(block));
        } else {
          if (node.argument.type !== 'Identifier') {
            scope.push({ id: id });
            path.insertBefore(t.expressionStatement(t.assignmentExpression('=', id, node.argument)));
          }
          var ret = t.returnStatement(id);
          ret.isTypeChecked = true;
          path.replaceWith(thrower({
            check: check,
            ret: ret,
            message: returnTypeErrorMessage(path, fn, id)
          }));
        }
      },
      VariableDeclaration: function VariableDeclaration(path) {
        var node = path.node;
        var scope = path.scope;

        var collected = [];
        var declarations = path.get("declarations");
        for (var i = 0; i < node.declarations.length; i++) {
          var declaration = node.declarations[i];
          var id = declaration.id;
          var init = declaration.init;

          if (!id.typeAnnotation || id.hasBeenTypeChecked) {
            continue;
          }
          id.savedTypeAnnotation = id.typeAnnotation;
          id.hasBeenTypeChecked = true;
          var ok = staticCheckAnnotation(declarations[i], id.typeAnnotation);
          if (ok === true) {
            continue;
          } else if (ok === false) {
            throw path.buildCodeFrameError('Invalid assignment value, expected ' + humanReadableType(id.typeAnnotation, scope));
          }
          var check = checkAnnotation(id, id.typeAnnotation, scope);
          if (check) {
            collected.push(guard({
              check: check,
              message: varTypeErrorMessage(id, scope)
            }));
          }
        }
        if (collected.length > 0) {
          var check = collected.reduce(function (check, branch) {
            branch.alternate = check;
            return branch;
          });
          if (path.parent.type === 'Program' || path.parent.type === 'BlockStatement') {
            path.insertAfter(check);
          } else {
            path.replaceWith(t.blockStatement([node, check]));
          }
        }
      },
      AssignmentExpression: function AssignmentExpression(path) {
        var node = path.node;
        var scope = path.scope;

        if (node.hasBeenTypeChecked || node.left.hasBeenTypeChecked) {
          return;
        }
        var binding = scope.getBinding(node.left.name);
        if (!binding || binding.path.type !== 'VariableDeclarator') {
          return;
        }
        var annotation = path.get('left').getTypeAnnotation();
        if (annotation.type === 'AnyTypeAnnotation') {
          var item = binding.path.get('id');
          annotation = item.node.savedTypeAnnotation || item.getTypeAnnotation();
        }

        node.hasBeenTypeChecked = true;
        node.left.hasBeenTypeChecked = true;
        if (annotation.type === 'AnyTypeAnnotation') {
          annotation = getAnnotation(path.get('right'));
          if (allowsAny(annotation)) {
            return;
          }
        }
        var id = node.left;
        var right = path.get('right');
        var ok = staticCheckAnnotation(right, annotation);
        if (ok === true) {
          return;
        } else if (ok === false) {
          throw path.buildCodeFrameError('Invalid assignment value, expected ' + humanReadableType(annotation, scope));
        }
        var check = checkAnnotation(id, annotation, scope);
        if (!id.typeAnnotation) {
          id.typeAnnotation = annotation;
        }
        id.hasBeenTypeChecked = true;
        if (check) {
          path.getStatementParent().insertAfter(guard({
            check: check,
            message: varTypeErrorMessage(id, scope)
          }));
        }
      },
      TypeCastExpression: function TypeCastExpression(path) {
        var node = path.node;

        var target = undefined;
        switch (node.expression.type) {
          case 'Identifier':
            target = node.expression;
            break;
          case 'AssignmentExpression':
            target = node.expression.left;
            break;
          default:
            // unsupported.
            return;
        }
        var id = path.scope.getBindingIdentifier(target.name);
        if (!id) {
          return;
        }
        id.savedTypeAnnotation = path.getTypeAnnotation();
      }
    }
  };

  function createChecks() {
    return {
      number: template('typeof input === \'number\''),
      boolean: template('typeof input === \'boolean\''),
      function: template('typeof input === \'function\''),
      string: template('typeof input === \'string\''),
      symbol: template('typeof input === \'symbol\''),
      undefined: template('input === undefined'),
      null: template('input === null'),
      nullOrUndefined: template('input == null'),
      instanceof: template('input instanceof type'),
      type: template('type(input)'),
      mixed: function mixed() {
        return null;
      },
      any: function any() {
        return template('input != null').expression;
      },
      union: checkUnion,
      intersection: checkIntersection,
      array: checkArray,
      map: checkMap,
      set: checkSet,
      tuple: checkTuple,
      object: checkObject,
      nullable: checkNullable,
      typeof: checkTypeof
    };
  }

  function createStaticChecks() {
    return {
      string: function string(path) {
        return maybeStringAnnotation(getAnnotation(path));
      },
      number: function number(path) {
        return maybeNumberAnnotation(getAnnotation(path));
      },
      boolean: function boolean(path) {
        return maybeBooleanAnnotation(getAnnotation(path));
      },
      function: function _function(path) {
        return maybeFunctionAnnotation(getAnnotation(path));
      },
      any: function any(path) {
        return null;
      },
      instanceof: function _instanceof(_ref6) {
        var path = _ref6.path;
        var type = _ref6.type;
        var node = path.node;
        var scope = path.scope;

        if (type.name === 'Object' && !scope.hasBinding('Object') && node.type === 'ObjectExpression') {
          return true;
        }
        return maybeInstanceOfAnnotation(getAnnotation(path), type);
      },
      type: (function (_type) {
        function type(_x) {
          return _type.apply(this, arguments);
        }

        type.toString = function () {
          return _type.toString();
        };

        return type;
      })(function (_ref7) {
        var path = _ref7.path;
        var type = _ref7.type;

        return null;
      }),
      array: function array(_ref8) {
        var path = _ref8.path;
        var types = _ref8.types;
        var node = path.node;

        if (node.type === 'ArrayExpression') {
          if (types.length === 0) {
            return true;
          } else if (node.elements.length < types.length) {
            return false;
          } else {
            return null;
          }
        } else if (maybeArrayAnnotation(getAnnotation(path)) === false) {
          return false;
        } else {
          return null;
        }
      },
      tuple: function tuple(_ref9) {
        var path = _ref9.path;
        var types = _ref9.types;
        var node = path.node;

        var annotation = getAnnotation(path);
        if (annotation.type === 'TypeAnnotation' || annotation.type === 'NullableTypeAnnotation') {
          annotation = annotation.typeAnnotation;
        }
        return maybeTupleAnnotation(annotation);
      },

      union: checkStaticUnion,
      intersection: checkStaticIntersection,
      object: checkStaticObject,
      nullable: checkStaticNullable
    };
  }

  function compareAnnotations(a, b) {

    if (a.type === 'TypeAnnotation') {
      a = a.typeAnnotation;
    }
    if (b.type === 'TypeAnnotation') {
      b = b.typeAnnotation;
    }
    switch (a.type) {
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
        return maybeStringAnnotation(b);
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
        return maybeNumberAnnotation(b);
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
        return maybeBooleanAnnotation(b);
      case 'FunctionTypeAnnotation':
        return maybeFunctionAnnotation(b);
      case 'AnyTypeAnnotation':
        return null;
      case 'MixedTypeAnnotation':
        return null;
      case 'ObjectTypeAnnotation':
        return compareObjectAnnotation(a, b);
      case 'ArrayTypeAnnotation':
        return compareArrayAnnotation(a, b);
      case 'GenericTypeAnnotation':
        return compareGenericAnnotation(a, b);
      case 'TupleTypeAnnotation':
        return compareTupleAnnotation(a, b);
      case 'UnionTypeAnnotation':
        return compareUnionAnnotation(a, b);
      case 'IntersectionTypeAnnotation':
        return compareIntersectionAnnotation(a, b);
      case 'NullableTypeAnnotation':
        return compareNullableAnnotation(a, b);
      default:
        return null;
    }
  }

  function unionComparer(a, b, comparator) {
    var falseCount = 0;
    var trueCount = 0;
    if (!a.types) {
      //console.trace(a.type);
      return null;
    }
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = a.types[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var _type2 = _step.value;

        var result = comparator(_type2, b);
        if (result === true) {
          if (b.type !== 'UnionTypeAnnotation') {
            return true;
          }
          trueCount++;
        } else if (result === false) {
          if (b.type === 'UnionTypeAnnotation') {
            return false;
          }
          falseCount++;
        }
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    if (falseCount === a.types.length) {
      return false;
    } else if (trueCount === a.types.length) {
      return true;
    } else {
      return null;
    }
  }

  function intersectionComparer(a, b, comparator) {
    var falseCount = 0;
    var trueCount = 0;
    if (!a.types) {
      return null;
    }
    var _iteratorNormalCompletion2 = true;
    var _didIteratorError2 = false;
    var _iteratorError2 = undefined;

    try {
      for (var _iterator2 = a.types[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
        var _type3 = _step2.value;

        var result = comparator(_type3, b);
        if (result === true) {
          trueCount++;
        } else if (result === false) {
          return false;
        }
      }
    } catch (err) {
      _didIteratorError2 = true;
      _iteratorError2 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion2 && _iterator2.return) {
          _iterator2.return();
        }
      } finally {
        if (_didIteratorError2) {
          throw _iteratorError2;
        }
      }
    }

    if (trueCount === a.types.length) {
      return true;
    } else {
      return null;
    }
  }

  function compareObjectAnnotation(a, b) {
    switch (b.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return compareObjectAnnotation(a, b.typeAnnotation);
      case 'UnionTypeAnnotation':
        return unionComparer(a, b, compareObjectAnnotation);
      case 'IntersectionTypeAnnotation':
        return intersectionComparer(a, b, compareObjectAnnotation);
      case 'VoidTypeAnnotation':
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
      case 'FunctionTypeAnnotation':
        return false;
      default:
        return null;
    }
  }

  function compareArrayAnnotation(a, b) {
    switch (b.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return compareArrayAnnotation(a, b.typeAnnotation);
      case 'UnionTypeAnnotation':
        return unionComparer(a, b, compareArrayAnnotation);
      case 'IntersectionTypeAnnotation':
        return intersectionComparer(a, b, compareArrayAnnotation);
      case 'VoidTypeAnnotation':
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
      case 'FunctionTypeAnnotation':
        return false;
      default:
        return null;
    }
  }

  function compareGenericAnnotation(a, b) {
    switch (b.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return compareGenericAnnotation(a, b.typeAnnotation);
      case 'GenericTypeAnnotation':
        if (b.id.name === a.id.name) {
          return true;
        } else {
          return null;
        }
      case 'UnionTypeAnnotation':
        return unionComparer(a, b, compareGenericAnnotation);
      case 'IntersectionTypeAnnotation':
        return intersectionComparer(a, b, compareGenericAnnotation);
      default:
        return null;
    }
  }

  function compareTupleAnnotation(a, b) {
    if (b.type === 'TupleTypeAnnotation') {
      if (b.types.length === 0) {
        return null;
      } else if (b.types.length < a.types.length) {
        return false;
      }
      return a.types.every(function (type, index) {
        return compareAnnotations(type, b.types[index]);
      });
    }
    switch (b.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return compareTupleAnnotation(a, b.typeAnnotation);
      case 'UnionTypeAnnotation':
        return unionComparer(a, b, compareTupleAnnotation);
      case 'IntersectionTypeAnnotation':
        return intersectionComparer(a, b, compareTupleAnnotation);
      case 'VoidTypeAnnotation':
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
      case 'FunctionTypeAnnotation':
        return false;
      default:
        return null;
    }
  }

  function compareUnionAnnotation(a, b) {
    switch (b.type) {
      case 'NullableTypeAnnotation':
        return compareUnionAnnotation(a, b.typeAnnotation);
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
        return null;
      default:
        return unionComparer(a, b, compareAnnotations);
    }
  }

  function compareNullableAnnotation(a, b) {
    switch (b.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return compareNullableAnnotation(a, b.typeAnnotation);
      case 'NullableTypeAnnotation':
      case 'VoidTypeAnnotation':
        return null;
    }
    if (compareAnnotations(a.typeAnnotation, b) === true) {
      return true;
    } else {
      return null;
    }
  }

  function arrayExpressionToTupleAnnotation(path) {
    var elements = path.get('elements');
    return t.tupleTypeAnnotation(elements.map(function (element) {
      return getAnnotation(element);
    }));
  }

  function checkStaticUnion(_ref10) {
    var path = _ref10.path;
    var types = _ref10.types;

    var falseCount = 0;
    var nullCount = 0;
    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = types[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        var _type4 = _step3.value;

        var result = staticCheckAnnotation(path, _type4);
        if (result === true) {
          return true;
        } else if (result === false) {
          falseCount++;
        } else {
          nullCount++;
        }
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3.return) {
          _iterator3.return();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }

    if (falseCount === types.length) {
      return false;
    } else {
      return null;
    }
  }

  function checkStaticIntersection(_ref11) {
    var path = _ref11.path;
    var types = _ref11.types;

    var trueCount = 0;
    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
      for (var _iterator4 = types[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
        var _type5 = _step4.value;

        var result = staticCheckAnnotation(path, _type5);
        if (result === true) {
          trueCount++;
        } else if (result === false) {
          return false;
        }
      }
    } catch (err) {
      _didIteratorError4 = true;
      _iteratorError4 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion4 && _iterator4.return) {
          _iterator4.return();
        }
      } finally {
        if (_didIteratorError4) {
          throw _iteratorError4;
        }
      }
    }

    if (trueCount === types.length) {
      return true;
    } else {
      return null;
    }
  }

  function checkStaticObject(_ref12) {
    var path = _ref12.path;
    var type = _ref12.type;
  }

  function checkStaticNullable(_ref13) {
    var path = _ref13.path;
    var type = _ref13.type;

    var annotation = getAnnotation(path);
    if (annotation.type === 'VoidTypeAnnotation' || annotation.type === 'NullableTypeAnnotation') {
      return true;
    } else {
      return staticCheckAnnotation(path, type);
    }
  }

  function checkNullable(_ref14) {
    var input = _ref14.input;
    var type = _ref14.type;
    var scope = _ref14.scope;

    var check = checkAnnotation(input, type, scope);
    if (!check) {
      return;
    }
    return t.logicalExpression("||", checks.undefined({ input: input }).expression, check);
  }

  function checkTypeof(_ref15) {
    var input = _ref15.input;
    var annotation = _ref15.annotation;
    var scope = _ref15.scope;

    switch (annotation.type) {
      case 'GenericTypeAnnotation':
        var id = annotation.id;

        var path = { type: id.type, node: id, scope: scope };
        return checkAnnotation(input, getAnnotation(path), scope);
      default:
        return checkAnnotation(input, annotation, scope);
    }
  }

  function checkUnion(_ref16) {
    var input = _ref16.input;
    var types = _ref16.types;
    var scope = _ref16.scope;

    var checks = types.map(function (type) {
      return checkAnnotation(input, type, scope);
    }).filter(identity);
    return checks.reduce(function (last, check, index) {
      if (last == null) {
        return check;
      }
      return t.logicalExpression("||", last, check);
    }, null);
  }

  function checkIntersection(_ref17) {
    var input = _ref17.input;
    var types = _ref17.types;
    var scope = _ref17.scope;

    var checks = types.map(function (type) {
      return checkAnnotation(input, type, scope);
    }).filter(identity);
    return checks.reduce(function (last, check, index) {
      if (last == null) {
        return check;
      }
      return t.logicalExpression("&&", last, check);
    }, null);
  }

  function checkMap(_ref18) {
    var input = _ref18.input;
    var types = _ref18.types;
    var scope = _ref18.scope;

    var _types = _slicedToArray(types, 2);

    var keyType = _types[0];
    var valueType = _types[1];

    var key = t.identifier('key');
    var value = t.identifier('value');
    var keyCheck = keyType ? checkAnnotation(key, keyType, scope) : null;
    var valueCheck = valueType ? checkAnnotation(value, valueType, scope) : null;
    if (!keyCheck) {
      if (!valueCheck) {
        return checkIsMap({ input: input }).expression;
      } else {
        return checkMapValues({ input: input, value: value, valueCheck: valueCheck }).expression;
      }
    } else {
      if (!valueCheck) {
        return checkMapKeys({ input: input, key: key, keyCheck: keyCheck }).expression;
      } else {
        return checkMapEntries({ input: input, key: key, value: value, keyCheck: keyCheck, valueCheck: valueCheck }).expression;
      }
    }
  }

  function checkSet(_ref19) {
    var input = _ref19.input;
    var types = _ref19.types;
    var scope = _ref19.scope;

    var _types2 = _slicedToArray(types, 1);

    var valueType = _types2[0];

    var value = t.identifier('value');
    var valueCheck = valueType ? checkAnnotation(value, valueType, scope) : null;
    if (!valueCheck) {
      return checkIsSet({ input: input }).expression;
    } else {
      return checkSetEntries({ input: input, value: value, valueCheck: valueCheck }).expression;
    }
  }
  function checkArray(_ref20) {
    var input = _ref20.input;
    var types = _ref20.types;
    var scope = _ref20.scope;

    if (types.length === 0) {
      return checkIsArray({ input: input }).expression;
    } else if (types.length === 1) {
      var item = t.identifier('item');
      var _type6 = types[0];
      var check = checkAnnotation(item, _type6, scope);
      if (!check) {
        return checkIsArray({ input: input }).expression;
      }
      return t.logicalExpression('&&', checkIsArray({ input: input }).expression, t.callExpression(t.memberExpression(input, t.identifier('every')), [t.functionExpression(null, [item], t.blockStatement([t.returnStatement(check)]))]));
    } else {
      // This is a tuple
      var _checks = types.map(function (type, index) {
        return checkAnnotation(t.memberExpression(input, t.numericLiteral(index), true), type, scope);
      }).filter(identity);

      var checkLength = t.binaryExpression('>=', t.memberExpression(input, t.identifier('length')), t.numericLiteral(types.length));

      return _checks.reduce(function (last, check, index) {
        return t.logicalExpression("&&", last, check);
      }, t.logicalExpression('&&', checkIsArray({ input: input }).expression, checkLength));
    }
  }

  function checkTuple(_ref21) {
    var input = _ref21.input;
    var types = _ref21.types;
    var scope = _ref21.scope;

    if (types.length === 0) {
      return checkIsArray({ input: input }).expression;
    }

    // This is a tuple
    var checks = types.map(function (type, index) {
      return checkAnnotation(t.memberExpression(input, t.numericLiteral(index), true), type, scope);
    }).filter(identity);

    var checkLength = t.binaryExpression('>=', t.memberExpression(input, t.identifier('length')), t.numericLiteral(types.length));

    return checks.reduce(function (last, check, index) {
      return t.logicalExpression("&&", last, check);
    }, t.logicalExpression('&&', checkIsArray({ input: input }).expression, checkLength));
  }

  function checkObject(_ref22) {
    var input = _ref22.input;
    var properties = _ref22.properties;
    var scope = _ref22.scope;

    return properties.reduce(function (expr, prop, index) {
      var target = t.memberExpression(input, prop.key);
      var check = checkAnnotation(target, prop.value, scope);
      if (check) {
        if (prop.optional) {
          check = t.logicalExpression('||', checks.undefined({ input: target }).expression, check);
        }
        return t.logicalExpression("&&", expr, check);
      } else {
        return expr;
      }
    }, checkIsObject({ input: input }).expression);
  }

  function createTypeAliasChecks(path) {
    var node = path.node;
    var scope = path.scope;
    var id = node.id;
    var annotation = node.right;

    var input = t.identifier('input');
    var check = checkAnnotation(input, annotation, scope) || t.booleanLiteral(true);
    var declaration = declareTypeChecker({ id: id, check: check });
    declaration.isTypeChecker = true;
    return declaration;
  }

  function createInterfaceChecks(path) {
    var node = path.node;
    var scope = path.scope;
    var id = node.id;
    var annotation = node.body;

    var input = t.identifier('input');
    var check = node.extends.reduce(function (check, extender) {
      return t.logicalExpression('&&', check, checkAnnotation(input, t.genericTypeAnnotation(extender.id), path.scope));
      return check;
    }, checkAnnotation(input, annotation, scope) || t.booleanLiteral(true));

    var declaration = declareTypeChecker({ id: id, check: check });
    declaration.isTypeChecker = true;
    return declaration;
  }

  function checkAnnotation(input, annotation, scope) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return checkAnnotation(input, annotation.typeAnnotation, scope);
      case 'TypeofTypeAnnotation':
        return checks.typeof({ input: input, annotation: annotation.argument, scope: scope });
      case 'GenericTypeAnnotation':
        if (annotation.id.name === 'Array') {
          return checks.array({ input: input, types: annotation.typeParameters ? annotation.typeParameters.params : [], scope: scope });
        } else if (annotation.id.name === 'Map' && !scope.hasBinding('Map')) {
          return checks.map({ input: input, types: annotation.typeParameters ? annotation.typeParameters.params : [], scope: scope });
        } else if (annotation.id.name === 'Set' && !scope.hasBinding('Set')) {
          return checks.set({ input: input, types: annotation.typeParameters ? annotation.typeParameters.params : [], scope: scope });
        } else if (annotation.id.name === 'Function') {
          return checks.function({ input: input }).expression;
        } else if (isTypeChecker(annotation.id, scope)) {
          return checks.type({ input: input, type: annotation.id }).expression;
        } else if (isPolymorphicType(annotation.id, scope)) {
          return;
        } else {
          return checks.instanceof({ input: input, type: createTypeExpression(annotation.id) }).expression;
        }
      case 'TupleTypeAnnotation':
        return checks.tuple({ input: input, types: annotation.types, scope: scope });
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
        return checks.number({ input: input }).expression;
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
        return checks.boolean({ input: input }).expression;
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
        return checks.string({ input: input }).expression;
      case 'UnionTypeAnnotation':
        return checks.union({ input: input, types: annotation.types, scope: scope });
      case 'IntersectionTypeAnnotation':
        return checks.intersection({ input: input, types: annotation.types, scope: scope });
      case 'ObjectTypeAnnotation':
        return checks.object({ input: input, properties: annotation.properties, indexers: annotation.indexers, scope: scope });
      case 'ArrayTypeAnnotation':
        return checks.array({ input: input, types: [annotation.elementType], scope: scope });
      case 'FunctionTypeAnnotation':
        return checks.function({ input: input, params: annotation.params, returnType: annotation.returnType });
      case 'MixedTypeAnnotation':
        return checks.mixed({ input: input });
      case 'AnyTypeAnnotation':
      case 'ExistentialTypeParam':
        return checks.any({ input: input });
      case 'NullableTypeAnnotation':
        return checks.nullable({ input: input, type: annotation.typeAnnotation, scope: scope }).expression;
      case 'VoidTypeAnnotation':
        return checks.undefined({ input: input }).expression;
    }
  }

  function staticCheckAnnotation(path, annotation) {
    var other = getAnnotation(path);

    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return staticCheckAnnotation(path, annotation.typeAnnotation);
      case 'GenericTypeAnnotation':
        if (isTypeChecker(annotation.id, path.scope)) {
          return staticChecks.type({ path: path, type: annotation.id });
        } else if (isPolymorphicType(annotation.id, path.scope)) {
          return;
        } else {
          return staticChecks.instanceof({ path: path, type: createTypeExpression(annotation.id) });
        }
    }

    return compareAnnotations(annotation, other);
  }

  /**
   * Get the type annotation for a given node.
   */
  function getAnnotation(path) {
    var annotation = getAnnotationShallow(path);
    while (annotation && annotation.type === 'TypeAnnotation') {
      annotation = annotation.typeAnnotation;
    }
    return annotation || t.anyTypeAnnotation();
  }

  function getAnnotationShallow(path) {
    var node = path.node;
    var scope = path.scope;

    if (node.type === 'TypeAlias') {
      return node.right;
    } else if (!node.typeAnnotation && !node.savedTypeAnnotation && !node.returnType) {
      switch (path.type) {
        case 'Identifier':
          var id = scope.getBindingIdentifier(node.name);
          if (!id) {
            break;
          }
          if (id.savedTypeAnnotation) {
            return id.savedTypeAnnotation;
          } else if (id.returnType) {
            return id.returnType;
          } else if (id.typeAnnotation) {
            return id.typeAnnotation;
          } else if (isPolymorphicType(id, scope)) {
            return t.anyTypeAnnotation();
          } else {
            var binding = scope.getBinding(node.name);
            var violation = getConstantViolationsBefore(binding, path).pop();
            if (violation) {
              return getAnnotation(violation);
            }
          }
          return path.getTypeAnnotation();
        case 'NumericLiteral':
        case 'StringLiteral':
        case 'BooleanLiteral':
          return path.getTypeAnnotation();
        case 'CallExpression':
          var callee = path.get('callee');
          if (callee.type === 'Identifier') {
            var fn = getFunctionForIdentifier(callee);
            if (fn) {
              return getAnnotation(fn);
            }
          }
          break;
        case 'AssignmentExpression':
          return getAssignmentExpressionAnnotation(path);
        case 'MemberExpression':
          return getMemberExpressionAnnotation(path);
        case 'ArrayExpression':
          return getArrayExpressionAnnotation(path);
        case 'ObjectExpression':
          return getObjectExpressionAnnotation(path);
        case 'BinaryExpression':
          return getBinaryExpressionAnnotation(path);
        case 'BinaryExpression':
          return getBinaryExpressionAnnotation(path);
        case 'LogicalExpression':
          return getLogicalExpressionAnnotation(path);
        case 'ConditionalExpression':
          return getConditionalExpressionAnnotation(path);
        default:
          return path.getTypeAnnotation();

      }
    }
    return node.savedTypeAnnotation || node.returnType || node.typeAnnotation || path.getTypeAnnotation();
  }

  function getObjectMethodAnnotation(path) {
    var node = path.node;

    return t.functionTypeAnnotation(null, node.params.map(function (param) {
      return param.savedTypeAnnotation || param.typeAnnotation;
    }), null, node.savedTypeAnnotation || node.returnType || node.typeAnnotation || t.anyTypeAnnotation());
  }

  function getAssignmentExpressionAnnotation(path) {
    if (path.node.operator === '=') {
      return getAnnotation(path.get('right'));
    }
  }

  function getBinaryExpressionAnnotation(path) {
    var node = path.node;

    if (isBooleanExpression(node)) {
      return t.booleanTypeAnnotation();
    } else {
      return t.anyTypeAnnotation();
    }
  }

  function getLogicalExpressionAnnotation(path) {
    var node = path.node;

    if (isBooleanExpression(node)) {
      return t.booleanTypeAnnotation();
    } else {
      var left = path.get('left');
      var right = path.get('right');
      switch (node.operator) {
        case '&&':
        case '||':
          var _ref23 = [getAnnotation(left), getAnnotation(right)];
          left = _ref23[0];
          right = _ref23[1];

          if (t.isUnionTypeAnnotation(left)) {
            if (t.isUnionTypeAnnotation(right)) {
              return t.unionTypeAnnotation(left.types.concat(right.types));
            } else {
              return t.unionTypeAnnotation(left.types.concat(right));
            }
          } else {
            return t.unionTypeAnnotation([left, right]);
          }
      }
      return t.anyTypeAnnotation();
    }
  }

  function getConditionalExpressionAnnotation(path) {
    var node = path.node;

    var consequent = getAnnotation(path.get('consequent'));
    var alternate = getAnnotation(path.get('alternate'));
    if (t.isUnionTypeAnnotation(consequent)) {
      if (t.isUnionTypeAnnotation(alternate)) {
        return t.unionTypeAnnotation(consequent.types.concat(alternate.types));
      } else {
        return t.unionTypeAnnotation(consequent.types.concat(alternate));
      }
    } else {
      return t.unionTypeAnnotation([consequent, alternate]);
    }
  }

  function getArrayExpressionAnnotation(path) {
    return t.genericTypeAnnotation(t.identifier('Array'), path.get('elements').map(getAnnotation));
  }

  function getObjectExpressionAnnotation(path) {
    var annotation = t.objectTypeAnnotation(path.get('properties').map(function (property) {
      if (property.computed) {
        return;
      }
      switch (property.type) {
        case 'ObjectMethod':
          return t.objectTypeProperty(t.identifier(property.node.key.name), getObjectMethodAnnotation(property));
        case 'ObjectProperty':
          return t.objectTypeProperty(t.identifier(property.node.key.name), property.node.value.savedTypeAnnotation || property.node.value.typeAnnotation || t.anyTypeAnnotation());
      }
    }).filter(identity));
    return annotation;
  }

  function getMemberExpressionAnnotation(path) {
    if (path.node.computed) {
      return getComputedMemberExpressionAnnotation(path);
    }
    var object = path.get('object');

    var _path$get = path.get('property');

    var id = _path$get.node;
    var name = id.name;

    var annotation = getAnnotation(object);
    if (annotation.type === 'NullableTypeAnnotation') {
      annotation = annotation.typeAnnotation;
    }
    if (annotation.type === 'GenericTypeAnnotation') {
      var target = path.scope.getBinding(annotation.id.name);
      if (target) {
        annotation = getAnnotation(target.path);
      }
    }
    switch (annotation.type) {
      case 'ObjectTypeAnnotation':
        var _iteratorNormalCompletion5 = true;
        var _didIteratorError5 = false;
        var _iteratorError5 = undefined;

        try {
          for (var _iterator5 = annotation.properties[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
            var _step5$value = _step5.value;
            var key = _step5$value.key;
            var value = _step5$value.value;

            if (key.name === id.name) {
              return value;
            }
          }
        } catch (err) {
          _didIteratorError5 = true;
          _iteratorError5 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion5 && _iterator5.return) {
              _iterator5.return();
            }
          } finally {
            if (_didIteratorError5) {
              throw _iteratorError5;
            }
          }
        }

        break;
    }
    return path.getTypeAnnotation();
  }

  function getComputedMemberExpressionAnnotation(path) {
    var object = path.get('object');
    var property = path.get('property');
    var objectAnnotation = getAnnotation(object);
    if (objectAnnotation.type === 'TypeAnnotation' || objectAnnotation.type === 'NullableTypeAnnotation') {
      objectAnnotation = objectAnnotation.typeAnnotation;
    }
    var propertyAnnotation = getAnnotation(property);
    if (propertyAnnotation.type === 'TypeAnnotation' || propertyAnnotation.type === 'NullableTypeAnnotation') {
      propertyAnnotation = propertyAnnotation.typeAnnotation;
    }

    var _property$evaluate = property.evaluate();

    var confident = _property$evaluate.confident;
    var value = _property$evaluate.value;

    if (!confident) {
      return path.getTypeAnnotation();
    }
    switch (objectAnnotation.type) {
      case 'TupleTypeAnnotation':
        if (objectAnnotation.types.length === 0) {
          break;
        } else if (typeof value === 'number') {
          if (!objectAnnotation.types[value]) {
            throw path.buildCodeFrameError('Invalid computed member expression for tuple: ' + humanReadableType(objectAnnotation, path.scope));
          }
          return objectAnnotation.types[value];
        } else {
          throw path.buildCodeFrameError('Invalid computed member expression for tuple: ' + humanReadableType(objectAnnotation, path.scope));
        }
        break;
    }
    return path.getTypeAnnotation();
  }

  function getFunctionForIdentifier(path) {
    if (path.type !== 'Identifier') {
      return false;
    }
    var ref = path.scope.getBinding(path.node.name);
    if (!ref) {
      return false;
    }
    return t.isFunction(ref.path.parent) && ref.path.parentPath;
  }

  /**
   * Returns `true` if the annotation is definitely for an array,
   * otherwise `false`.
   */
  function isStrictlyArrayAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return isStrictlyArrayAnnotation(annotation.typeAnnotation);
      case 'GenericTypeAnnotation':
        return annotation.id.name === 'Array';
      case 'UnionTypeAnnotation':
        return annotation.types.every(isStrictlyArrayAnnotation);
      default:
        return false;
    }
  }

  function compareMaybeUnion(annotation, comparator) {
    var falseCount = 0;
    var _iteratorNormalCompletion6 = true;
    var _didIteratorError6 = false;
    var _iteratorError6 = undefined;

    try {
      for (var _iterator6 = annotation.types[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
        var _type7 = _step6.value;

        var result = comparator(_type7);
        if (result === true) {
          return true;
        } else if (result === false) {
          falseCount++;
        }
      }
    } catch (err) {
      _didIteratorError6 = true;
      _iteratorError6 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion6 && _iterator6.return) {
          _iterator6.return();
        }
      } finally {
        if (_didIteratorError6) {
          throw _iteratorError6;
        }
      }
    }

    if (falseCount === annotation.types.length) {
      return false;
    } else {
      return null;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with a number,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeNumberAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeNumberAnnotation(annotation.typeAnnotation);
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
        return true;
      case 'GenericTypeAnnotation':
        switch (annotation.id.name) {
          case 'Array':
          case 'Function':
          case 'Object':
          case 'String':
          case 'Boolean':
          case 'Date':
          case 'RegExp':
            return false;
          default:
            return null;
        }
      case 'UnionTypeAnnotation':
        return compareMaybeUnion(annotation, maybeNumberAnnotation);
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with a string,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeStringAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeStringAnnotation(annotation.typeAnnotation);
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
        return true;
      case 'GenericTypeAnnotation':
        switch (annotation.id.name) {
          case 'Array':
          case 'Function':
          case 'Object':
          case 'Number':
          case 'Boolean':
          case 'Date':
          case 'RegExp':
            return false;
          default:
            return null;
        }
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion7 = true;
        var _didIteratorError7 = false;
        var _iteratorError7 = undefined;

        try {
          for (var _iterator7 = annotation.types[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
            var _type8 = _step7.value;

            var result = maybeStringAnnotation(_type8);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError7 = true;
          _iteratorError7 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion7 && _iterator7.return) {
              _iterator7.return();
            }
          } finally {
            if (_didIteratorError7) {
              throw _iteratorError7;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with a boolean,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeBooleanAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeBooleanAnnotation(annotation.typeAnnotation);
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
        return true;
      case 'GenericTypeAnnotation':
        switch (annotation.id.name) {
          case 'Array':
          case 'Function':
          case 'Object':
          case 'String':
          case 'Number':
          case 'Date':
          case 'RegExp':
            return false;
          default:
            return null;
        }
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion8 = true;
        var _didIteratorError8 = false;
        var _iteratorError8 = undefined;

        try {
          for (var _iterator8 = annotation.types[Symbol.iterator](), _step8; !(_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done); _iteratorNormalCompletion8 = true) {
            var _type9 = _step8.value;

            var result = maybeBooleanAnnotation(_type9);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError8 = true;
          _iteratorError8 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion8 && _iterator8.return) {
              _iterator8.return();
            }
          } finally {
            if (_didIteratorError8) {
              throw _iteratorError8;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with a function,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeFunctionAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeFunctionAnnotation(annotation.typeAnnotation);
      case 'FunctionTypeAnnotation':
        return true;
      case 'GenericTypeAnnotation':
        switch (annotation.id.name) {
          case 'Array':
          case 'Number':
          case 'Object':
          case 'String':
          case 'Boolean':
          case 'Date':
          case 'RegExp':
            return false;
          default:
            return null;
        }
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion9 = true;
        var _didIteratorError9 = false;
        var _iteratorError9 = undefined;

        try {
          for (var _iterator9 = annotation.types[Symbol.iterator](), _step9; !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
            var _type10 = _step9.value;

            var result = maybeFunctionAnnotation(_type10);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError9 = true;
          _iteratorError9 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion9 && _iterator9.return) {
              _iterator9.return();
            }
          } finally {
            if (_didIteratorError9) {
              throw _iteratorError9;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with an undefined or null type,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeNullableAnnotation(annotation) {
    switch (annotation.type) {
      case 'NullableTypeAnnotation':
      case 'VoidTypeAnnotation':
      case 'MixedTypeAnnotation':
        return true;
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return maybeNullableAnnotation(annotation.typeAnnotation);
      case 'GenericTypeAnnotation':
        switch (annotation.id.name) {
          case 'Array':
          case 'Number':
          case 'Object':
          case 'String':
          case 'Boolean':
          case 'Date':
          case 'RegExp':
            return false;
          default:
            return null;
        }
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion10 = true;
        var _didIteratorError10 = false;
        var _iteratorError10 = undefined;

        try {
          for (var _iterator10 = annotation.types[Symbol.iterator](), _step10; !(_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done); _iteratorNormalCompletion10 = true) {
            var _type11 = _step10.value;

            var result = maybeNullableAnnotation(_type11);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError10 = true;
          _iteratorError10 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion10 && _iterator10.return) {
              _iterator10.return();
            }
          } finally {
            if (_didIteratorError10) {
              throw _iteratorError10;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with an object type,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeInstanceOfAnnotation(annotation, expected) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeInstanceOfAnnotation(annotation.typeAnnotation);
      case 'GenericTypeAnnotation':
        if (annotation.id.name === expected.name) {
          return true;
        } else {
          return null;
        }
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion11 = true;
        var _didIteratorError11 = false;
        var _iteratorError11 = undefined;

        try {
          for (var _iterator11 = annotation.types[Symbol.iterator](), _step11; !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
            var _type12 = _step11.value;

            var result = maybeInstanceOfAnnotation(_type12, expected);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError11 = true;
          _iteratorError11 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion11 && _iterator11.return) {
              _iterator11.return();
            }
          } finally {
            if (_didIteratorError11) {
              throw _iteratorError11;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'VoidTypeAnnotation':
      case 'BooleanTypeAnnotation':
      case 'BooleanLiteralTypeAnnotation':
      case 'StringTypeAnnotation':
      case 'StringLiteralTypeAnnotation':
      case 'NumberTypeAnnotation':
      case 'NumericLiteralTypeAnnotation':
      case 'FunctionTypeAnnotation':
        return false;
      default:
        return null;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with an array,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeArrayAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeArrayAnnotation(annotation.typeAnnotation);
      case 'TupleTypeAnnotation':
      case 'ArrayTypeAnnotation':
        return true;
      case 'GenericTypeAnnotation':
        return annotation.id.name === 'Array' ? true : null;
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion12 = true;
        var _didIteratorError12 = false;
        var _iteratorError12 = undefined;

        try {
          for (var _iterator12 = annotation.types[Symbol.iterator](), _step12; !(_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done); _iteratorNormalCompletion12 = true) {
            var _type13 = _step12.value;

            var result = maybeArrayAnnotation(_type13);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError12 = true;
          _iteratorError12 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion12 && _iterator12.return) {
              _iterator12.return();
            }
          } finally {
            if (_didIteratorError12) {
              throw _iteratorError12;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'AnyTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  /**
   * Returns `true` if the annotation is compatible with a tuple,
   * `false` if it definitely isn't, or `null` if we're not sure.
   */
  function maybeTupleAnnotation(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
      case 'NullableTypeAnnotation':
        return maybeTupleAnnotation(annotation.typeAnnotation);
      case 'TupleTypeAnnotation':
        return true;
      case 'UnionTypeAnnotation':
        var falseCount = 0;
        var _iteratorNormalCompletion13 = true;
        var _didIteratorError13 = false;
        var _iteratorError13 = undefined;

        try {
          for (var _iterator13 = annotation.types[Symbol.iterator](), _step13; !(_iteratorNormalCompletion13 = (_step13 = _iterator13.next()).done); _iteratorNormalCompletion13 = true) {
            var _type14 = _step13.value;

            var result = maybeTupleAnnotation(_type14);
            if (result === true) {
              return true;
            } else if (result === false) {
              falseCount++;
            }
          }
        } catch (err) {
          _didIteratorError13 = true;
          _iteratorError13 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion13 && _iterator13.return) {
              _iterator13.return();
            }
          } finally {
            if (_didIteratorError13) {
              throw _iteratorError13;
            }
          }
        }

        if (falseCount === annotation.types.length) {
          return false;
        } else {
          return null;
        }
      case 'GenericTypeAnnotation':
      case 'AnyTypeAnnotation':
      case 'ArrayTypeAnnotation':
      case 'MixedTypeAnnotation':
      case 'IntersectionTypeAnnotation':
        return null;
      default:
        return false;
    }
  }

  function humanReadableType(annotation) {
    switch (annotation.type) {
      case 'TypeAnnotation':
      case 'FunctonTypeParam':
        return humanReadableType(annotation.typeAnnotation);

      case 'FunctionTypeAnnotation':
        // @fixme babel doesn't seem to like generating FunctionTypeAnnotations yet
        return '(' + annotation.params.map(humanReadableType).join(', ') + ') => ' + humanReadableType(annotation.returnType);
      default:
        return (0, _babelGenerator2.default)(annotation).code;
    }
  }

  function humanReadableObject(annotation, scope) {
    if (annotation.properties.length === 0) {
      return 'an object';
    } else {
      var shape = (0, _babelGenerator2.default)(annotation).code;
      return 'an object with shape ' + shape;
    }
  }

  function humanReadableArray(annotation, scope) {
    return (0, _babelGenerator2.default)(annotation).code;
  }

  function humanReadableTuple(annotation, scope) {
    return (0, _babelGenerator2.default)(annotation).code;
  }

  function isTypeChecker(id, scope) {
    var binding = scope.getBinding(id.name);
    if (binding === undefined) {
      return false;
    }
    var path = binding.path;

    return path != null && (path.type === 'TypeAlias' || path.type === 'ImportSpecifier' || path.type === 'VariableDeclaration' && path.node.isTypeChecker);
  }

  function isPolymorphicType(id, scope) {
    var binding = scope.getBinding(id.name);
    if (binding !== undefined) {
      return false;
    }
    var path = scope.path;

    while (path && path.type !== 'Program') {
      var _path = path;
      var node = _path.node;

      if (t.isFunction(node) && node.typeParameters) {
        var _iteratorNormalCompletion14 = true;
        var _didIteratorError14 = false;
        var _iteratorError14 = undefined;

        try {
          for (var _iterator14 = node.typeParameters.params[Symbol.iterator](), _step14; !(_iteratorNormalCompletion14 = (_step14 = _iterator14.next()).done); _iteratorNormalCompletion14 = true) {
            var param = _step14.value;

            param.isPolymorphicType = true;
            if (param.name === id.name) {
              return true;
            }
          }
        } catch (err) {
          _didIteratorError14 = true;
          _iteratorError14 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion14 && _iterator14.return) {
              _iterator14.return();
            }
          } finally {
            if (_didIteratorError14) {
              throw _iteratorError14;
            }
          }
        }
      }
      path = path.parent;
    }
    return false;
  }

  function getPolymorphicType(id, scope) {
    var binding = scope.getBinding(id.name);
    if (binding !== undefined) {
      return false;
    }
    var path = scope.path;

    while (path && path.type !== 'Program') {
      var _path2 = path;
      var node = _path2.node;

      if (t.isFunction(node) && node.typeParameters) {
        var _iteratorNormalCompletion15 = true;
        var _didIteratorError15 = false;
        var _iteratorError15 = undefined;

        try {
          for (var _iterator15 = node.typeParameters.params[Symbol.iterator](), _step15; !(_iteratorNormalCompletion15 = (_step15 = _iterator15.next()).done); _iteratorNormalCompletion15 = true) {
            var param = _step15.value;

            param.isPolymorphicType = true;
            if (param.name === id.name) {
              return param;
            }
          }
        } catch (err) {
          _didIteratorError15 = true;
          _iteratorError15 = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion15 && _iterator15.return) {
              _iterator15.return();
            }
          } finally {
            if (_didIteratorError15) {
              throw _iteratorError15;
            }
          }
        }
      }
      path = path.parent;
    }
    return null;
  }

  function collectParamChecks(path) {
    return path.get('params').map(function (param) {
      var node = param.node;

      if (node.type === 'AssignmentPattern') {
        if (node.left.typeAnnotation) {
          return createDefaultParamGuard(param);
        }
      } else if (node.type === 'RestElement') {
        if (node.typeAnnotation) {
          return createRestParamGuard(param);
        }
      } else if (node.typeAnnotation) {
        return createParamGuard(param);
      }
    }).filter(identity);
  }

  function createParamGuard(path) {
    var node = path.node;
    var scope = path.scope;

    node.hasBeenTypeChecked = true;
    node.savedTypeAnnotation = node.typeAnnotation;
    var check = checkAnnotation(node, node.typeAnnotation, scope);
    if (!check) {
      return;
    }
    if (node.optional) {
      check = t.logicalExpression('||', checks.undefined({ input: node }).expression, check);
    }
    var message = paramTypeErrorMessage(node, scope);
    return guard({
      check: check,
      message: message
    });
  }

  /**
   * Get any constant violations before a given node.
   * @fixme this is a copy of the internal babel api and relies on a private method.
   */
  function getConstantViolationsBefore(binding, path, functions) {
    var violations = binding.constantViolations.slice();
    violations.unshift(binding.path);
    return violations.filter(function (violation) {
      violation = violation.resolve();
      var status = violation._guessExecutionStatusRelativeTo(path);
      if (functions && status === "function") {
        functions.push(violation);
      }
      return status === "before";
    });
  }

  function createDefaultParamGuard(path) {
    var node = path.node;
    var scope = path.scope;
    var id = node.left;
    var value = node.right;

    var ok = staticCheckAnnotation(path.get('right'), id.typeAnnotation);
    if (ok === false) {
      throw path.buildCodeFrameError('Invalid default value for argument "' + id.name + '", expected ' + humanReadableType(id.typeAnnotation, scope) + '.');
    }
    return createParamGuard(path.get('left'));
  }

  function createRestParamGuard(path) {
    var node = path.node;
    var scope = path.scope;
    var id = node.argument;

    id.hasBeenTypeChecked = true;
    node.savedTypeAnnotation = node.typeAnnotation;
    if (!isStrictlyArrayAnnotation(node.typeAnnotation)) {
      throw path.buildCodeFrameError('Invalid type annotation for rest argument "' + id.name + '", expected an Array, got: ' + humanReadableType(node.typeAnnotation, scope) + '.');
    }
    var check = checkAnnotation(id, node.typeAnnotation, scope);
    if (!check) {
      return;
    }
    if (node.optional) {
      check = t.logicalExpression('||', checks.undefined({ input: id }).expression, check);
    }
    var message = paramTypeErrorMessage(id, scope, node.typeAnnotation);
    return guard({
      check: check,
      message: message
    });
  }

  function returnTypeErrorMessage(path, fn, id) {
    var node = path.node;
    var scope = path.scope;

    var name = fn.id ? fn.id.name : '';
    var message = 'Function ' + (name ? '"' + name + '" ' : '') + 'return value violates contract, expected ' + humanReadableType(fn.returnType, scope) + ' got ';

    return t.binaryExpression('+', t.stringLiteral(message), node.argument ? readableName({ input: id || node.argument }).expression : t.stringLiteral('undefined'));
  }

  function paramTypeErrorMessage(node, scope) {
    var typeAnnotation = arguments.length <= 2 || arguments[2] === undefined ? node.typeAnnotation : arguments[2];

    var name = node.name;
    var message = 'Value of ' + (node.optional ? 'optional ' : '') + 'argument "' + name + '" violates contract, expected ' + humanReadableType(typeAnnotation, scope) + ' got ';

    return t.binaryExpression('+', t.stringLiteral(message), readableName({ input: node }).expression);
  }

  function varTypeErrorMessage(node, scope, annotation) {
    var name = node.name;
    var message = 'Value of variable "' + name + '" violates contract, expected ' + humanReadableType(annotation || node.typeAnnotation, scope) + ' got ';
    return t.binaryExpression('+', t.stringLiteral(message), readableName({ input: node }).expression);
  }

  /**
   * Determine whether the given node can produce purely boolean results.
   */
  function isBooleanExpression(node) {
    if (node.type === 'BinaryExpression' && BOOLEAN_BINARY_OPERATORS.indexOf(node.operator) > -1) {
      return true;
    } else if (node.type === 'LogicalExpression') {
      return isBooleanExpression(node.left) && isBooleanExpression(node.right);
    } else {
      return false;
    }
  }

  /**
   * Convert type specifier to expression.
   */
  function createTypeExpression(node) {
    if (node.type == 'Identifier') {
      return node;
    } else if (node.type == 'QualifiedTypeIdentifier') {
      return t.memberExpression(createTypeExpression(node.qualification), createTypeExpression(node.id));
    }

    throw this.errorWithNode('Unsupported type: ' + node.type);
  }

  /**
   * Get name of a type as a string.
   */
  function getTypeName(node) {
    if (node.type == 'Identifier') {
      return node.name;
    } else if (node.type == 'QualifiedTypeIdentifier') {
      return getTypeName(node.qualification) + '.' + getTypeName(node.id);
    }

    throw this.errorWithNode('Unsupported type: ' + node.type);
  }

  /**
   * Union two arrays.
   */
  function union(arr1, arr2) {
    for (var i = 0; i < arr2.length; i++) {
      var item = arr2[i];
      if (arr1.indexOf(item) === -1) {
        arr1.push(item);
      }
    }
    return arr1;
  }

  /**
   * Determine whether the given annotation allows any value.
   */
  function allowsAny(annotation) {
    if (annotation.type === 'TypeAnnotation' || annotation.type === 'NullableTypeAnnotation') {
      return allowsAny(annotation.typeAnnotation);
    } else if (annotation.type === 'AnyTypeAnnotation' || annotation.type === 'MixedTypeAnnotation') {
      return true;
    } else if (annotation.type === 'UnionTypeAnnotation') {
      return annotation.types.some(allowsAny);
    } else {
      return false;
    }
  }

  /**
   * Determine whether a given node is nully (null or undefined).
   */
  function isNodeNully(node) {
    if (node == null) {
      return true;
    } else if (node.type === 'Identifier' && node.name === 'undefined') {
      return true;
    } else if (node.type === 'Literal' && node.value === null) {
      return true;
    } else if (node.type === 'UnaryExpression' && node.operator === 'void') {
      return true;
    } else {
      return false;
    }
  }

  /**
   * A function that returns its first argument, useful when filtering.
   */
  function identity(input) {
    return input;
  }
};

var _babelGenerator = require('babel-generator');

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }